-- ("*****         RESOURCE ALLOCATION WITH JAVA IMPLICIT UNIQUE CONDITION");
-- ("******           JAVA STANDARD POLICY");
with Ada.Text_IO;  use Ada.Text_IO;
package body Allocation is
	Max : constant := 100; 
	type Cardinal is new Integer range 0..2;
	--  WaitCount : Integer := 0; -- for counting number of waiting steps 

	protected Controler is
		entry Request(Y : Integer; Score : in out Cardinal);
		procedure Release(Y : Integer);
		-- Score records the allocation steps of the request
	private
		entry Wait(Y : Integer; Score : in out Cardinal);
		Stock : Integer := Max;
		Requested : Boolean := False;
		NotifyAll : Boolean := False; -- Java signal simulation
	end Controler;

	protected body Controler is
		entry Request(Y : Integer; Score : in out Cardinal) when True is
		begin
			case Score is
				when 0 => if not Requested then
							Requested := True; 
							Stock := Stock - Y;
							if Stock < 0 then 
								Score := 1; 
							else 
								Score := 2; Requested := False;
							end if;
						  end if;
				when 1 => if Stock >= 0 then 
							Score := 2; Requested := False;
						  end if;
				when 2 => null;
			end case;
			if Score /= 2 then 
				NotifyAll := False; requeue Wait; -- Java wait simulation
			end if;
		end Request;

     	entry Wait(Y : Integer; Score : in out Cardinal) when NotifyAll is
      	begin
			null; -- all queued tasks quit the monitor
			WaitCount:= WaitCount + 1;
      	end Wait;

		procedure Release(Y : Integer) is
		begin
			Stock := Stock + Y;
			If Stock >= 0 then NotifyAll := True; end if;
		end Release;
	end Controler;

   	procedure Request(Y : Integer; Caller : Id) is		
		Score :  Cardinal := 0;
		-- Score records the allocation steps of the request 
	begin
		-- the following participate simulating the Java signalling semantics	
		-- * Put_Line("     New Client " & Id'Image(Caller) & "  Requesting " & Integer'Image(Y));
		while Score /= 2 loop 
			Controler.Request(Y, Score); 
			--  * Put_Line("Client " & Id'Image(Caller) & "  Requesting " & Integer'Image(Y)
			--*		 & " Score = " & Cardinal'Image(Score));
		end loop;
	end Request;
	
	procedure Release(Y : Integer; Caller : Id) is
	begin 
		Controler.Release(Y); 
		--* Put_Line("    Client " & Id'Image(Caller) & "  has released " & Integer'Image(Y));
	end Release;

begin
	New_Line;
	Put_Line("*****  RESOURCE ALLOCATION WITH JAVA IMPLICIT UNIQUE CONDITION");
	Put_Line("******  JAVA STANDARD POLICY");
	New_Line;
	Put_Line("****** " & Integer'Image(Global.N) & " Clients, " & Integer'Image(Global.R) & " cycles. That is "
						& Integer'Image(Global.N * Global.R) & " Requests");
end Allocation;

