-- allocation policy implemented with a monitor and 2 conditions 
-- ("*****         RESOURCE ALLOCATION WITH JAVA TWO EXPLICIT CONDITION");
-- ("******           JAVA STANDARD POLICY OF J2SE/JDK 1.5"); 

-- JAVA implementation with Lock and conditions of J2SE/JDK 1.5
-- condition synchronisation provided by additional condition objects
-- Solution when a monitor allows several named condition variable

-- simulated with an Ada protected object 
--   which simulates Java concurrency semantics
-- the Java condition.await operation is here an entry call with a barrier

with Ada.Text_IO;  use Ada.Text_IO;

package body Allocation is
	Max : constant := 100; 
	subtype Cardinal is Integer range 0..2;
	--* type SticScore is array(Id) of Cardinal;
	--* Score : SticScore := (others => False);
	
	protected Controler is
		entry Request(Y : Integer; Score : in out Cardinal);
		procedure Release(Y : Integer);
		-- Score records the allocation steps of the request
	private
		entry Participation(Y : Integer; Score : in out Cardinal);
		entry Grant(Y : Integer; Score : in out Cardinal);
		Stock : Integer := Max;
		Requested : Boolean := False;
		SignalParticipation : Boolean :=  False; -- Java signals simulation  
		SignalGrant : Boolean := False; -- Java signals simulation
	end Controler;
	
    protected body Controler is
   
   		entry Request(Y : Integer; Score : in out Cardinal) when True is
		begin
			case Score is
				when 0 => if not Requested then
							Requested := True; 
							Stock := Stock - Y;
							if Stock < 0 then 
								Score := 1; SignalGrant:= false; requeue Grant;
							else 
								Score := 2; Requested := False; SignalParticipation := True;
							end if;
						  else
							SignalParticipation := False; requeue Participation;
						  end if;
				when 1 => if Stock >= 0 then 
							Score := 2; Requested := False; SignalParticipation := True;
						  end if;
				when 2 => null;  -- back to the Java ready thread queue
			end case; 
		end Request;
       
      entry Participation(Y : Integer; Score : in out Cardinal) when SignalParticipation is
      begin
			null; SignalParticipation := False;
			ParticipationCount := ParticipationCount + 1;
      end Participation; -- back to the Java ready thread queue 

      entry Grant(Y : Integer; Score : in out Cardinal) when SignalGrant is
      begin
			null; GrantCount := GrantCount + 1;
      end Grant; -- back to the Java ready thread queue

     procedure Release(Y : Integer) is
      begin
		Stock := Stock + Y;
		If Stock >= 0 then SignalGrant := True; end if;
	  end Release; -- back to the Java ready thread queue 
	  
    end Controler;
 	
   	procedure Request(Y : Integer; Caller : Id) is		
		Score :  Cardinal := 0;
		-- Score records the allocation steps of the request 
	begin
		-- the following participate simulating the Java signalling semantics	
		--* Put_Line("     New Client " & Id'Image(Caller) & "  Requesting " & Integer'Image(Y));
		while Score /= 2 loop 
			Controler.Request(Y, Score); 
		--*	 Put_Line("Client " & Id'Image(Caller) & "  Requesting " & Integer'Image(Y)
		--*			 & " Score = " & Cardinal'Image(Score));
		end loop;
	end Request;
	
	procedure Release(Y : Integer; Caller : Id) is
	begin 
		Controler.Release(Y); 
	--*	 Put_Line("    Client " & Id'Image(Caller) & "  has released " & Integer'Image(Y));
	end Release;

begin
	New_Line;
	Put_Line("*****         RESOURCE ALLOCATION WITH JAVA TWO EXPLICIT CONDITION");
	Put_Line("******        JAVA STANDARD POLICY OF J2SE/JDK 1.5");
	New_Line;
	Put_Line("****** " & Integer'Image(Global.N) & " Clients, " & Integer'Image(Global.R) & " cycles. That is "
						& Integer'Image(Global.N * Global.R) & " Requests");
end Allocation;
