-- MALL JAVA PO version simulating Java with Ada monitor semantics ----------------------

-- this implementation is safe due to monitor semantic 
-- giving priority to awaked signalled processes over new callers

-- this version of the mall is implemented with a protected object Cooperation_Synchro

-- the Java wait operation is here an entry call with a barrier named NotifyAll

-- the Java synchronized method is simulated by Cooperation_Synchro
-- Score is an in out parameter indicating the stage reached by the caller
-- Score = 0 is the first request
-- Score = 1 when the first partner has registered and waits for the arrival of the second
-- Score = 2 when the partner Id is returned

with Text_IO; use Text_IO;
package body Mall is

   type Cardinal is new Integer range 0..2;
  --  WaitCount : Integer := 0; -- for counting number of waiting steps 

   protected Cooperation_Synchro is
      entry Cooperate(X: in Id_Chameneos; X_Other: out Id_Chameneos; Score : in out Cardinal);
   private
      entry Wait(X: in Id_Chameneos; X_Other: out Id_Chameneos; Score : in out Cardinal);
      First_Call : Boolean := True;
	  MustWait : Boolean := False;
      A_Chameneos   : Id_Chameneos;
      B_Chameneos   : Id_Chameneos;
	  NotifyAll : Boolean := False; -- Java signal simulation
   end Cooperation_Synchro;

   protected body Cooperation_Synchro is
      entry Cooperate(X: in Id_Chameneos; X_Other: out Id_Chameneos; Score : in out Cardinal) when not NotifyAll is
      begin
		--* if (Score = 0) and (MustWait) then
			-- Java defensive programming forbidding a newcommer to overcome the waiting partner
		--*	NotifyAll := False; requeue Wait; -- Java wait simulation
		--* end if;	
         if (Score = 0) and (First_Call) then
			-- this is the action of the first partner of the pair
            A_Chameneos := X;
            First_Call := False; 
			Score := 1; -- the first partner has registered and waits for the arrival of the second
			-- Put_line(Id_Chameneos'image(X) & "           IS WAITING IN MALL");
           --*  NotifyAll := False; 
			requeue Wait; -- Java wait simulation 
		 end if;
		 if (Score = 0) and (not First_Call) then
			-- this is the action of the second partner of the pair
            B_Chameneos := X;
            X_Other := A_Chameneos;
            First_Call := True;
			MustWait := True; -- the next user of the monitor must be the waiting partner
		 	Score := 2; --  the partner Id of the second partner is returned
			NotifyAll := True; -- awakes the first partner
         end if;
         if (Score = 1) and (First_Call) then
			X_Other := B_Chameneos;
			MustWait := False; -- each of the pair has been given the name of its partner
		 	Score := 2; -- this is the first partner again and its partner Id is returned
			-- * NotifyAll := True;
		 end if;
      end Cooperate;

      entry  Wait(X: in Id_Chameneos; X_Other: out Id_Chameneos; Score : in out Cardinal) when NotifyAll is
      begin
			null; WaitCount:= WaitCount + 1;
		    NotifyAll := False; 
			requeue Cooperate;
      end Wait;
   end Cooperation_Synchro;

   function  Get_A_Peer(X: Id_Chameneos) return Id_Chameneos is
      Other_Chameneos : Id_Chameneos;
	  Score :  Cardinal := 0;
	  -- Score is an in out parameter indicating the stage reached by the caller
   begin
	  while Score /= 2 loop
	  	Cooperation_Synchro.Cooperate(X, Other_Chameneos, Score);
		delay(0.001);
	  end loop;
      return Other_Chameneos;
   end Get_A_Peer;
begin
	New_Line;
	Put_Line("  *****      CHAMENEOS WITH JAVA PROPOSED MONITOR SIMULATION ");  
end Mall;
