-- dining philosophers paradigm.

-- Solution implementing global allocation (and instrumentation)
-- The chopsticks are allocated only when both are available  
 
-- ADA solution simulating the JAVA MONITOR PROPOSED A LA ADA SEMANTICS 
-- GIVING  PRIORITY TO AWAKED PROCESSES OVER NEW REQUESTORS
-- instrumented for concurrency denials ratio measurement
-- attempt to have a fair solution
-- requestor added 
-- no requestors cycling and causing deadlock
--  this avoids deadlock, it is fair and has a good concurrency ratio
-- due to the first test in Get_LR. this test is not necessary (it is reliable without it)
-- however it provides more concurrency and thus more efficiency 

-- ***************     reliable and fair,  deadlock avoided,  fair

with Global; use Global;
with Monitoring; use Monitoring;
with Ada.Text_IO;  use Ada.Text_IO;
package body Chop is

	type SticState is array(Id) of Boolean;
		
	protected Sticks is
		entry Get_LR(C : in Id; Done : out Boolean); 
		procedure Release(C : in Id);
	private 
		entry Wait(C : in Id; Done : out Boolean); 
		Available : SticState := (others => True); -- Stick availability
		Requestor : SticState := (others => False); -- booking philosophers
		StillToNotify : Integer := 0;
	    NotifyAll : Boolean := False; -- Java signal simulation
				
	end Sticks;
	
	protected body Sticks is   

 		entry Get_LR(C : in Id; Done : out Boolean) when not NotifyAll is
		begin
			if Requestor(C - 1) and Requestor(C + 1) then
				Monitoring.InstrumentGetRefusal(C); -- counts one more refusal
				Done:= False ;  requeue Wait;
			end if;
			if Available (C) and Available (C + 1) and not Requestor(C - 1)  then
				Available (C) := False; Available (C + 1) := False; Done := True;
			    Requestor(C) := False;
				Monitoring.InstrumentGetSuccessful(C); -- counts the success once
			else 
				Monitoring.InstrumentGetRefusal(C); -- counts one more refusal 
		    	Requestor(C) := True;
				Done := False;  requeue Wait;
			end if;						
		end Get_LR;    

 		entry Wait(C : in Id; Done : out Boolean) when NotifyAll is
		begin
			 -- moves from the waiting queue to the method Get_LR()  
			StillToNotify := StillToNotify - 1;
			NotifyAll := StillToNotify > 0; 
			-- when False, opens the Get_LR barrier allowing non queued requests
			requeue Get_LR;
		end Wait;    

		procedure Release(C : Id) is
		begin
			Available(C) := True; Available(C + 1) := True; 
			-- awakes now all blocked tasks as if Java's notifyAll
			StillToNotify := Wait'Count; 
			NotifyAll := StillToNotify > 0; 
			Monitoring.InstrumentRelease(C);
		end Release;
	end Sticks;

	procedure Get_LR(C : Id) is
		OK : Boolean := False;	Score : Integer;
	begin
		ChopStart(C); -- starting service time for sticks allocation
		while not OK loop 
			Sticks.Get_LR(C, OK); 
			if OK then Score := 2; else Score := 0; end if;
			Requesting_LR(C, Score);
		end loop;
		ChopEnd(C); -- ending  service time for sticks allocation
	end Get_LR;

	procedure Release(C : Id) is begin Sticks.Release(C); end Release;

begin
	New_Line;
	Put_Line("*****   DINING PHILOSOPHERS WITH STICK GLOBALLY ALLOCATED ");
	Put_Line("*****   ADA SIMULATION OF JAVA MONITOR PROPOSED PRIORITY SEMANTICS ");
	Put_Line("*****   RELIABLE AND FAIR,  USING REQUESTOR for fairness ");
	Put_Line("*****   Good concurrency ration thanks to a preliminary test when requesting") ;
	Put_Line("*****   WITH CONCURRENCY RATIO INSTRUMENTATION");
end Chop;
