-- dining philosophers paradigm.

-- Solution implementing global allocation (and instrumentation)
-- The chopsticks are allocated only when both are available  
-- RELIABLE BUT UNFAIR

-- ADA solution using the eggshell semantic of Ada protected object 
-- instrumented for concurrency denials ratio measurement
-- a simulation shows how unfair it can be (3% of accesses for a starving philosopher)

-- ***************     reliable, no deadlock, not fair, starvation prone

with Global; use Global;
with Monitoring; use Monitoring;
with Ada.Text_IO;  use Ada.Text_IO;
package body Chop is

	type SticState is array(Id) of Boolean;
		
	protected Sticks is
		entry Get_LR(C : in Id); 
		procedure Release(C : in Id);
	private 
		entry Assign_LR(C : in Id); 
		Available : SticState := (others => True); -- Stick availability
	    Flush_Count : Natural := 0;
				
	end Sticks;
	
	protected body Sticks is   

 		entry Get_LR(C : in Id) when True is
		begin
			if Available (C) and Available (C + 1) then
				Available (C) := False; Available (C + 1) := False;
				Monitoring.InstrumentGetSuccessful(C); -- counts the success once
			else 
				Monitoring.InstrumentGetRefusal(C); -- counts one more refusal 
				requeue Assign_LR;
			end if;						
		end Get_LR;    

 		entry Assign_LR(C : in Id) when Flush_Count > 0 is
		begin
			Flush_Count := Flush_Count - 1;			
			if Available (C) and Available (C + 1) then
				Available (C) := False; Available (C + 1) := False;
				Monitoring.InstrumentGetSuccessful(C);  -- counts the success once
 			else 
				Monitoring.InstrumentGetRefusal(C); -- counts one more refusal each time
				requeue Assign_LR;
			end if;
		end Assign_LR;    

		procedure Release(C : Id) is
		begin
			Available(C) := True; Available(C + 1) := True; 
			Flush_Count := Assign_LR'Count;
			Monitoring.InstrumentRelease(C);
		end Release;
	end Sticks;

	procedure Get_LR(C : Id) is	
	begin
		ChopStart(C); -- starting service time for sticks allocation 
		Sticks.Get_LR(C); 
		ChopEnd(C); -- ending  service time for sticks allocation
	end Get_LR;

	procedure Release(C : Id) is begin Sticks.Release(C); end Release;

begin
	New_Line;
	Put_Line("*****   DINING PHILOSOPHERS WITH GLOBAL ALLOCATION ");
	Put_Line("*****   ADA PROTECTED OBJECT SOLUTION : RELIABLE, UNFAIR");
	Put_Line("*****   WITH CONCURRENCY RATIO INSTRUMENTATION");
end Chop;
