-- dining philosophers paradigm.

-- Solution implementing global allocation (and instrumentation)
-- The chopsticks are allocated only when both are available  
 
-- ADA solution simulating the Java monitor semantics
-- instrumented for concurrency denials ratio measurement
-- attempt to have a fair solution
-- requestor added but this introduces deadlock

-- ***************     unreliable and fair,  deadlock caused by fearness attempt, starvation prone

with Global; use Global;
with Monitoring; use Monitoring;
with Ada.Text_IO;  use Ada.Text_IO;
package body Chop is

	type SticState is array(Id) of Boolean;
		
	protected Sticks is
		entry Get_LR(C : in Id; Done : out Boolean); 
		procedure Release(C : in Id);
	private 
		entry Wait(C : in Id; Done : out Boolean); 
		Available : SticState := (others => True); -- Stick availability
		Requestor : SticState := (others => False); -- booking philosophers
	    NotifyAll : Boolean := False; -- Java signal simulation
				
	end Sticks;
	
	protected body Sticks is   

 		entry Get_LR(C : in Id; Done : out Boolean) when True is
		begin
			if Available (C) and Available (C + 1) and not Requestor(C - 1) and not Requestor(C+1)  then
				Available (C) := False; Available (C + 1) := False; Done := True;
			    Requestor(C) := False;
				Monitoring.InstrumentGetSuccessful(C); -- counts the success once
			else 
				Monitoring.InstrumentGetRefusal(C); -- counts one more refusal 
		    	Requestor(C) := True;
				Done := False; NotifyAll := False; requeue Wait;
			end if;						
		end Get_LR;    

 		entry Wait(C : in Id; Done : out Boolean) when NotifyAll is
		begin
			null; -- moves from the waiting queue to the ready queue 
		end Wait;    

		procedure Release(C : Id) is
		begin
			Available(C) := True; Available(C + 1) := True; 
			NotifyAll := (Wait'Count > 0); -- Java NotyfyAll simulation
			Monitoring.InstrumentRelease(C);
		end Release;
	end Sticks;

	procedure Get_LR(C : Id) is
		OK : Boolean := False;	Score : Integer;
	begin
		ChopStart(C); -- starting service time for sticks allocation
		while not OK loop 
			Sticks.Get_LR(C, OK); 
			if OK then Score := 2; else Score := 0; end if;
			Requesting_LR(C, Score);
		end loop;
		ChopEnd(C); -- ending  service time for sticks allocation
	end Get_LR;

	procedure Release(C : Id) is begin Sticks.Release(C); end Release;

begin
	New_Line;
	Put_Line("*****   DINING PHILOSOPHERS WITH STICK GLOBALLY ALLOCATED ");
	Put_Line("*****   ADA SIMULATION OF JAVA MONITOR SOLUTION : UNRELIABLE due to fairness attempt");
	Put_Line("*****   WITH CONCURRENCY RATIO INSTRUMENTATION");
end Chop;
