import java.util.concurrent.locks.Condition ;
import java.util.concurrent.locks.Lock ;
import java.util.concurrent.locks.ReentrantLock ;

public final class Resource {

	private int Max ; // Number of resources
	private int stock = Max ; // Current number of non allocated resources
	private boolean requested = false;
	final Lock lock = new ReentrantLock();
	final Condition Participation = lock.newCondition() ; 
	final Condition Grant = lock.newCondition();


	Resource (int Max) {
		this.Max = Max ;
		this.stock = Max;
	}

	public void request(int number)  {
		lock.lock();                     // mutual exclusion entrance
		try {
			while ( requested ) Participation.await(); 
			requested = true;
			stock -= number; // new value of stock, may be negative
			System.out.println(" requesting " + number + "   leaving stock of " + stock);
			while (stock < 0) Grant.await() ;
			requested = false;	
			Participation.signal(); // for a possibly blocked request
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} finally {
 			lock.unlock();              // mutual exclusion leave
			System.out.println( " lock is released by request");
		}
	}

	public void release (int number) {
		lock.lock();                    // mutual exclusion entrance
		System.out.println( " lock is acquired by release");
		stock += number ;
		if (stock >= 0 ) Grant.signal() ;
		System.out.println(" releasing " + number + "   raising stock to " + stock);
		lock.unlock();                     // mutual exclusion leave
		System.out.println( " lock is released by release");
	}

}

// The first requesting thread is served first
// either it receives its requested number at first call
// or once signalled it will be the sole allowed to proceed when calling request anew
// other callers access is denied until the priviledged request is granted its full request
